//
//  GlkPairWindow.m
//  CocoaGlk
//
//  Created by Andrew Hunter on 19/03/2005.
//  Copyright 2005 Andrew Hunter. All rights reserved.
//

#import "GlkPairWindow.h"


@implementation GlkPairWindow

// = Initialisation =

- (id) init {
	self = [super init];
	
	if (self) {
		borderWidth = 2;
	}
	
	return self;
}

- (void) dealloc {
	[left setParent: nil];
	[right setParent: nil];
	
	[key release]; key = nil;
	[left release]; left = nil;
	[right release]; right = nil;
	
	[super dealloc];
}

// = Setting the windows that make up this pair =

- (void) setKeyWindow: (GlkWindow*) newKey {
	[key release];
	
	key = [newKey retain];
	
	needsLayout = YES;
}

- (void) setLeftWindow: (GlkWindow*) newLeft {
	[left setParent: nil];
	[left removeFromSuperview];
	[left release]; 

	left = [newLeft retain];
	[left setParent: self];
	
	needsLayout = YES;
}

- (void) setRightWindow: (GlkWindow*) newRight {
	[right setParent: nil];
	[right removeFromSuperview];
	[right release]; 
	
	right = [newRight retain];
	[right setParent: self];
	
	needsLayout = YES;
}

- (GlkWindow*) keyWindow {
	return key;
}

- (GlkWindow*) leftWindow {
	return left;
}

- (GlkWindow*) rightWindow {
	return right;
}

// = Size and arrangement =

- (void) setSize: (unsigned) newSize {
	size = newSize;
	
	needsLayout = YES;
}

- (void) setFixed: (BOOL) newFixed {
	fixed = newFixed;
	
	needsLayout = YES;
}

- (void) setHorizontal: (BOOL) newHorizontal {
	horizontal = newHorizontal;
	
	needsLayout = YES;
}

- (void) setAbove: (BOOL) newAbove {
	above = newAbove;
}

- (unsigned) size {
	return size;
}

- (BOOL) fixed {
	return fixed;
}

- (BOOL) horizontal {
	return horizontal;
}

- (BOOL) above {
	return above;
}

// = Custom settings =

- (void) setBorderWidth: (float) newBorderWidth {
	borderWidth = newBorderWidth;
	
	needsLayout = YES;
}

// = Layout =

- (void) layoutInRect: (NSRect) parentRect {
	if (needsLayout || !NSEqualRects(parentRect, [self frame])) {
		// Set our own frame
		[self setFrame: parentRect];
		
		NSRect bounds = [self bounds];
		
		// Work out the sizes for the child windows
		float availableSize = horizontal?parentRect.size.width:parentRect.size.height;
		availableSize -= borderWidth;
		
		float leftSize, rightSize;
		
		if (fixed) {
			if (horizontal) {
				rightSize = [right widthForFixedSize: size];
			} else {
				rightSize = [right heightForFixedSize: size];
			}
		} else {
			rightSize = (availableSize * ((float)size))/100.0;
		}
		
		if (rightSize > availableSize) rightSize = availableSize-1.0;

		rightSize = floorf(rightSize);		
		leftSize = floorf(availableSize - rightSize);
		
		NSRect leftRect;
		NSRect rightRect;
		
		if (horizontal) {
			borderSliver.size.width = borderWidth;
			
			if (above) {
				leftRect.origin.x = bounds.origin.x + rightSize + borderWidth;
				rightRect.origin.x = bounds.origin.x;
				borderSliver.origin.x = bounds.origin.x + rightSize;
			} else {
				leftRect.origin.x = bounds.origin.x;
				rightRect.origin.x = bounds.origin.x + leftSize + borderWidth;
				borderSliver.origin.x = bounds.origin.x + leftSize;
			}
			
			borderSliver.origin.y = leftRect.origin.y = rightRect.origin.y = bounds.origin.y;
			borderSliver.size.height = leftRect.size.height = rightRect.size.height = bounds.size.height;
			leftRect.size.width = leftSize;
			rightRect.size.width = rightSize;
		} else {
			borderSliver.size.height = borderWidth;
			
			if (!above) {
				leftRect.origin.y = bounds.origin.y + rightSize + borderWidth;
				rightRect.origin.y = bounds.origin.y;
				borderSliver.origin.y = bounds.origin.y + rightSize;
			} else {
				leftRect.origin.y = bounds.origin.y;
				rightRect.origin.y = bounds.origin.y + leftSize + borderWidth;
				borderSliver.origin.y = bounds.origin.y + leftSize;
			}
			
			borderSliver.origin.x = leftRect.origin.x = rightRect.origin.x = bounds.origin.x;
			borderSliver.size.width = leftRect.size.width = rightRect.size.width = bounds.size.width;
			leftRect.size.height = leftSize;
			rightRect.size.height = rightSize;
		}
		
		if ([left parent] != self) {
			NSLog(@"GlkPairWindow: left parent does not match self");
			return;
		}
		
		if ([right parent] != self) {
			NSLog(@"GlkPairWindow: right parent does not match self");
			return;
		}
		
		// Perform the layout
		if ([left superview] != self) {
			[left removeFromSuperview];
			[self addSubview: left];
		}
		
		if ([right superview] != self) {
			[right removeFromSuperview];
			[self addSubview: right];
		}
		
		[left layoutInRect: leftRect];
		[right layoutInRect: rightRect];
		
		[self setNeedsDisplay: YES];
		
		needsLayout = NO;
	} else {
		// Nothing major to do, but pass the buck anyway
		[left layoutInRect: [left frame]];
		[right layoutInRect: [right frame]];
	}
}

- (float) widthForFixedSize: (unsigned) sz {
	if (key && [key closed]) {
		[key release]; key = nil;
	}
	
	if (key) {
		return [key widthForFixedSize: sz];
	} else {
		return 0;
	}
}

- (float) heightForFixedSize: (unsigned) sz {
	if (key && [key closed]) {
		[key release]; key = nil;
	}
	
	if (key) {
		return [key heightForFixedSize: sz];
	} else {
		return 0;
	}
}

// = Window control =

- (void) setEventTarget: (NSObject<GlkEventReceiver>*) newTarget {
	[super setEventTarget: newTarget];
	
	// Propagate the handler
	[left setEventTarget: newTarget];
	[right setEventTarget: newTarget];
}

- (void) bufferIsFlushing {
	[super bufferIsFlushing];
	
	[left bufferIsFlushing];
	[right bufferIsFlushing];
}

- (void) bufferHasFlushed {
	[super bufferHasFlushed];
	
	[left bufferHasFlushed];
	[right bufferHasFlushed];
}

// = Window metadata =

- (void) setClosed: (BOOL) newClosed {
	[super setClosed: newClosed];
	
	// This propagates
	[left setClosed: newClosed];
	[right setClosed: newClosed];
}

// = Drawing =

- (void)drawRect:(NSRect)rect {
	[[NSColor windowBackgroundColor] set];
	NSRectFill(borderSliver);
	
	if (borderWidth >= 2) {
		NSRect bounds = [self bounds];
		
		if (!horizontal) {
			// Draw lines at top and bottom of border
			[[NSColor controlHighlightColor] set];
			NSRectFill(NSMakeRect(NSMinX(bounds), NSMinY(borderSliver), bounds.size.width, 1));
			
			[[NSColor controlShadowColor] set];
			NSRectFill(NSMakeRect(NSMinX(bounds), NSMaxY(borderSliver)-1, bounds.size.width, 1));
		} else {
			// Draw lines at left and right of border
			[[NSColor controlHighlightColor] set];
			NSRectFill(NSMakeRect(NSMinX(borderSliver), NSMinY(bounds), 1, bounds.size.height));
			
			[[NSColor controlShadowColor] set];
			NSRectFill(NSMakeRect(NSMaxX(borderSliver)-1, NSMinY(bounds), 1, bounds.size.height));
		}
	}
}

@end
