//
//  GlkTextGridWindow.m
//  CocoaGlk
//
//  Created by Andrew Hunter on 20/03/2005.
//  Copyright 2005 Andrew Hunter. All rights reserved.
//

#import "GlkTextGridWindow.h"


@implementation GlkTextGridWindow

// = Initialisation =

- (id)initWithFrame:(NSRect)frame {
    self = [super initWithFrame:frame];
    
	if (self) {
		xpos = 0;
		ypos = 0;
		lines = [[NSMutableArray alloc] init];
		
		[self setForceFixed: YES];
    }
	
    return self;
}

- (void) dealloc {
	[lines release];
	
	[super dealloc];
}

// = Drawing =

- (void)drawRect:(NSRect)rect {
	// Clear to the background colour
	[[self backgroundColour] set];
	NSRectFill(rect);
	
	// Draw the lines
	NSRect bounds = [self bounds];
	int y;
	float lineHeight = [self lineHeight];
	
	for (y=0; y<[lines count]; y++) {
		[(NSAttributedString*)[lines objectAtIndex: y] drawAtPoint: NSMakePoint(NSMinX(bounds), NSMaxY(bounds) - (y+1)*lineHeight)];
	}
}

// = Layout =

- (float) charWidth {
	// FIXME: we should cache this
	return [@"M" sizeWithAttributes: [self currentTextAttributes]].width;
}

- (float) widthForFixedSize: (unsigned) size {
	NSSize baseSize = [@"M" sizeWithAttributes: [self currentTextAttributes]];
	
	return floorf(size * baseSize.width);
}

- (float) heightForFixedSize: (unsigned) size {
	return floorf(size * [self lineHeight]);
}

- (GlkSize) glkSize {
	GlkSize res;
	
	res.width = width;
	res.height = height;
	
	return res;
}

- (void) layoutInRect: (NSRect) parentRect {
	// Set our frame
	[self setFrame: parentRect];
	
	// Lay out the lines
	width  = parentRect.size.width  / [self charWidth];
	height = parentRect.size.height / [self lineHeight];
	
	// Add new lines
	int x;
	
	for (x=[lines count]; x<height; x++) {
		[lines addObject: [[[NSMutableAttributedString alloc] init] autorelease]];
	}
	
	// Remove old lines
	if ([lines count] > height) {
		[lines removeObjectsInRange: NSMakeRange(height, [lines count]-height)];
	}
	
	// Size lines
	for (x=0; x<height; x++) {
		NSMutableAttributedString* line = [lines objectAtIndex: x];
		
		if ([line length] < width) {
			// Add spaces to the end (not sure about the attributes to use)
			int amountToAdd = width - [line length];
			unichar* spaces = malloc(sizeof(unichar)*amountToAdd);
			int y;
			
			for (y=0; y<amountToAdd; y++) {
				spaces[y] = ' ';
			}
			
			NSAttributedString* string = [[NSAttributedString alloc] initWithString: [NSString stringWithCharacters: spaces
																											 length: amountToAdd]
																		 attributes: [self currentTextAttributes]];
			
			[line appendAttributedString: string];
			[string release];
			free(spaces);
		} else if ([line length] > width) {
			[line deleteCharactersInRange: NSMakeRange(width, [line length]-width)];
		}
	}
}

// = Cursor positioning =

- (void) moveCursorToXposition: (int) newXpos
					 yPosition: (int) newYpos {
	xpos = newXpos;
	ypos = newYpos;
}

// = Window control =

- (void) clearWindow {
	[lines release]; lines = nil;
	lines = [[NSMutableArray alloc] init];
	
	xpos = ypos = 0;
	
	// This forces us to recreate the lines appropriately (ie, this works as if the window was shrunk to 0,0 then resized back to normal)
	[self layoutInRect: [self frame]];
}

// = Streams =

- (void) putString: (NSString*) string {
	int pos = 0;
	
	// Check for newlines
	int x;
	for (x=0; x<[string length]; x++) {
		if ([string characterAtIndex: x] == '\n' || [string characterAtIndex: x] == '\r') {
			[self putString: [string substringToIndex: x]];
			xpos = 0; ypos++;
			[self putString: [string substringFromIndex: x+1]];
			return;
		}
	}
	
	// Write this string
	while (pos < [string length]) {
		// Can't draw if we've fallen off the end of the window
		if (ypos >= [lines count] || ypos > height) 
			break;
		
		// Can only draw a certain number of characters
		if (xpos >= width) {
			xpos = 0;
			ypos++;
			continue;
		}
		
		// Get the number of characters to draw
		int amountToDraw = width - xpos;
		if (amountToDraw > [string length]) amountToDraw = [string length];
		
		// Draw the characters
		NSAttributedString* partString = [[NSAttributedString alloc] initWithString: [string substringWithRange: NSMakeRange(pos, amountToDraw)]
																		 attributes: [self currentTextAttributes]];
		[[lines objectAtIndex: ypos] replaceCharactersInRange: NSMakeRange(xpos, amountToDraw)
										withAttributedString: partString];
		[partString release];
		
		[self setNeedsDisplay: YES];
		
		// Update the x position (and the y position if necessary)
		xpos += amountToDraw;
		pos += amountToDraw;
		if (xpos >= width) {
			xpos = 0;
			ypos++;
		}
	}
}

// = Mouse input =

- (void) mouseDown: (NSEvent*) event {
	NSRect bounds = [self bounds];
	NSPoint mousePos = [self convertPoint: [event locationInWindow] 
								 fromView: nil];
	
	float charWidth = [self charWidth];
	float lineHeight = [self lineHeight];
	
	int clickX = mousePos.x / charWidth;
	int clickY = (NSMaxY(bounds)-mousePos.y) / lineHeight;
	
	if (mouseInput) {
		// Generate the event
		GlkEvent* evt = [[GlkEvent alloc] initWithType: evtype_MouseInput
									  windowIdentifier: [self identifier]
												  val1: clickX
												  val2: clickY];
		
		// ... send it
		[target queueEvent: [evt autorelease]];
	} else {
		[super mouseDown: event];
	}
}

@end
