//
//  GlkWindow.m
//  CocoaGlk
//
//  Created by Andrew Hunter on 19/03/2005.
//  Copyright 2005 Andrew Hunter. All rights reserved.
//

#import "GlkWindow.h"


@implementation GlkWindow

// = Initialisation =

- (id)initWithFrame:(NSRect)frame {
    self = [super initWithFrame:frame];
    
	if (self) {
		border = 4;
	}
    
	return self;
}

- (void) dealloc {
	[styles release]; styles = nil;
	[preferences release]; preferences = nil;
	
	[super dealloc];
}

// = Drawing =

- (void)drawRect:(NSRect)rect {
	[[self backgroundColour] set];
	NSRectFill(rect);
}

- (BOOL) isOpaque {
	return YES;
}

// = Window metadata =

- (void) setClosed: (BOOL) newClosed {
	closed = newClosed;
}

- (BOOL) closed {
	return closed;
}

- (void) setIdentifier: (unsigned) newWindowIdentifier {
	windowIdentifier = newWindowIdentifier;
}

- (unsigned) identifier {
	return windowIdentifier;
}

// = The parent window =

- (void) setParent: (GlkPairWindow*) newParent {
	parentWindow = newParent;
}

- (GlkPairWindow*) parent {
	return parentWindow;
}

// = Layout =

- (void) layoutInRect: (NSRect) parentRect {
	[self setFrame: parentRect];
}

- (float) widthForFixedSize: (unsigned) size {
	return size;
}

- (float) heightForFixedSize: (unsigned) size {
	return size;
}

- (void) setBorder: (float) newBorder {
	border = newBorder;
}

- (float) border {
	return border;
}

- (NSRect) contentRect {
	return NSInsetRect([self bounds], border, border);
}

- (GlkSize) glkSize {
	NSRect contentRect = [self contentRect];
	GlkSize res;
	
	res.width = contentRect.size.width;
	res.height = contentRect.size.height;
	
	return res;
}

// = Styles =

- (void) setForceFixed: (BOOL) newForceFixed {
	forceFixed = newForceFixed;
}

- (BOOL) forceFixed {
	return forceFixed;
}

- (NSColor*) backgroundColour {
	return [[self style: style_Normal] backColour];
}

- (NSFont*) proportionalFont {
	if (forceFixed) {
		return [self fixedFont];
	} else {
		return [[self attributes: style_Normal] objectForKey: NSFontAttributeName];
	}
}

- (NSFont*) fixedFont {
	return [[self attributes: style_Preformatted] objectForKey: NSFontAttributeName];
}

- (NSDictionary*) currentTextAttributes {
	return [self attributes: style];
}

- (float) leading {
	return 0;
}

- (float) lineHeight {
	return [[[self currentTextAttributes] objectForKey: NSFontAttributeName] defaultLineHeightForFont];
}

- (void) setStyles: (NSDictionary*) newStyles {
	[styles release];
	styles = [[NSDictionary alloc] initWithDictionary: newStyles
											copyItems: YES];
}

- (GlkStyle*) style: (unsigned) glkStyle {
	// If there aren't any styles yet, get the default styles from the preferences
	if (!styles) {
		if (!preferences) preferences = [[GlkPreferences sharedPreferences] retain];
		[self setStyles: [preferences styles]];
	}
	
	// Get the result from the styles object (use a default if we can't find a suitable style)
	GlkStyle* res = [styles objectForKey: [NSNumber numberWithUnsignedInt: glkStyle]];	
	if (!res) res = [GlkStyle style];
	
	if (forceFixed && [res proportional]) [res setProportional: NO];
	if (forceFixed && glkStyle != style_Normal) [res setSize: [[self style: style_Normal] size]];
	
	return res;
}

- (NSDictionary*) attributes: (unsigned) glkStyle {
	if (!preferences) preferences = [[GlkPreferences sharedPreferences] retain];
	
	GlkStyle* sty = [self style: glkStyle];
	
	return [sty attributesWithPreferences: preferences];
}

- (void) setPreferences: (GlkPreferences*) prefs {
	[preferences release];
	preferences = [prefs retain];
}

- (void) reformat {
	// Blank window just needs laying out again
	[self layoutInRect: [self frame]];
}

// = Cursor positioning =

- (void) moveCursorToXposition: (int) xpos
					 yPosition: (int) ypos {
	NSLog(@"Warning: attempt to move cursor in a window that doesn't support it");
}


// = Window control =

- (void) clearWindow {
	// We can't get any clearer
}

- (void) setEventTarget: (NSObject<GlkEventReceiver>*) newTarget {
	target = newTarget;
}

- (void) requestCharInput {
	if (lineInput) {
		NSLog(@"Oops: client requested char input while line input was pending");
		[self cancelLineInput];
	}
	charInput = YES;
	[[self window] invalidateCursorRectsForView: self];
}

- (void) requestLineInput {
	if (charInput) {
		NSLog(@"Oops: client requested line input while char input was pending");
		[self cancelCharInput];
	}
	lineInput = YES;
	[[self window] invalidateCursorRectsForView: self];
}

- (BOOL) waitingForKeyboardInput {
	return charInput || lineInput;
}

- (NSResponder*) windowResponder {
	return self;
}

- (void) setInputLine: (NSString*) inputLine {
	// As we don't support line input, there's nothing to do here
}

- (void) forceLineInput: (NSString*) forcedInput {
	// Can't deal with line input events, but character input events are easy
	if (charInput) {
		// Generate a character input event
		GlkEvent* glkEvent = [[GlkEvent alloc] initWithType: evtype_CharInput
										   windowIdentifier: [self identifier]
													   val1: [[self class] keycodeForString: forcedInput]
													   val2: 0];		
		[self cancelCharInput];
		[target queueEvent: [glkEvent autorelease]];
	}
}

- (void) requestMouseInput {
	mouseInput = YES;
	[[self window] invalidateCursorRectsForView: self];
}

- (void) cancelCharInput {
	charInput = NO;
	[[self window] invalidateCursorRectsForView: self];
}

- (void) cancelLineInput {
	lineInput = NO;
	[[self window] invalidateCursorRectsForView: self];
}

- (void) cancelMouseInput {
	mouseInput = NO;
	[[self window] invalidateCursorRectsForView: self];
}

// = Standard mouse and input handlers =

- (BOOL)acceptsFirstResponder {
	// Note that we can't handle line input events by default, so we only accept if we have character events
	if (charInput) {
		return YES;
	} else {
		return NO;
	}
}

+ (unsigned) keycodeForString: (NSString*) string {
	glui32 chr = keycode_Unknown;						// The Glk character
	
	if ([string length] <= 0) return chr;
	
	unichar inChar = [string characterAtIndex: 0];
	switch (inChar) {
		case '\n':
		case '\r':
			chr = keycode_Return;
			break;
		case '\t':
			chr = keycode_Tab;
			break;
			
		case NSUpArrowFunctionKey:
			chr = keycode_Up;
			break;
		case NSDownArrowFunctionKey:
			chr = keycode_Down;
			break;
		case NSLeftArrowFunctionKey:
			chr = keycode_Left;
			break;
		case NSRightArrowFunctionKey:
			chr = keycode_Right;
			break;
			
		case NSPageDownFunctionKey:
			chr = keycode_PageDown;
			break;
		case NSPageUpFunctionKey:
			chr = keycode_PageUp;
			break;
			
		case NSHomeFunctionKey:
			chr = keycode_Home;
			break;
			
		case NSEndFunctionKey:
			chr = keycode_End;
			break;
			
		case '\e':
			chr = keycode_Escape;
			break;
	}
	
	if (chr == keycode_Unknown) {
		NSData* latin1 = [string dataUsingEncoding: NSISOLatin1StringEncoding
							  allowLossyConversion: YES];
		
		if ([latin1 length] > 0) {
			chr = ((unsigned char*)[latin1 bytes])[0];
		}
	}
	
	return chr;
}

+ (unsigned) keycodeForEvent: (NSEvent*) evt {
	return [[self class] keycodeForString: [evt characters]];
}

- (void) keyDown: (NSEvent*) evt {
	if (!charInput) {
		NSBeep();
	} else if ([[evt characters] length] >= 1) {
		GlkEvent* glkEvent = [[GlkEvent alloc] initWithType: evtype_CharInput
										   windowIdentifier: [self identifier]
													   val1: [[self class] keycodeForEvent: evt]
													   val2: 0];
		
		[self cancelCharInput];
		[target queueEvent: [glkEvent autorelease]];
	}
}

- (int) inputPos {
	// Default is 0 (not managing a text view)
	return 0;
}

- (void) bufferIsFlushing {
	// Default action is to catch flies
}

- (void) bufferHasFlushed {
	// The horrible taste of flies fails to wake us up
}

// = Streaming =

// Control

- (void) closeStream {
	// Nothing to do really
}

- (void) setPosition: (in int) position
		  relativeTo: (in enum GlkSeekMode) seekMode {
	// No effect
}

- (unsigned) getPosition {
	// Spec isn't really clear on what do for window streams. We just say the position is always 0
	return 0;
}

// Writing

- (void) putChar: (unichar) ch {
	unichar buf[1];
	
	buf[0] = ch;
	
	[self putString: [NSString stringWithCharacters: buf
											 length: 1]];
}

- (void) putString: (NSString*) string {
	// We're blank: nothing to do
}

- (void) putBuffer: (NSData*) buffer {
	// Assume that buffers are in ISO Latin-1 format
	[self putString: [[[NSString alloc] initWithBytes: [buffer bytes]
											   length: [buffer length]
											 encoding: NSISOLatin1StringEncoding] autorelease]];
}

// = Reading =

- (unichar) getChar {
	return 0;
}

- (NSString*) getLineWithLength: (int) len {
	return nil;
}

- (NSData*) getBufferWithLength: (unsigned) length {
	return nil;
}

// Styles

- (void) setStyle: (int) styleId {
	style = styleId;
}

- (int) style {
	return style;
}

// = Cursor rects =

- (void)resetCursorRects {
	if (lineInput || charInput) {
		[self addCursorRect: [self bounds]
					 cursor: [NSCursor IBeamCursor]];
	} else if (mouseInput) {
		[self addCursorRect: [self bounds]
					 cursor: [NSCursor crosshairCursor]];
	} else {
	}
}


// = The containing view =

- (GlkView*) containingView {
	return containingView;
}

- (void) setContainingView: (GlkView*) view {
	containingView = view;
}

@end
